/*
 ***********************************************************************************************************************
 *
 *  Copyright (c) 2019-2024 Advanced Micro Devices, Inc. All Rights Reserved.
 *
 *  Permission is hereby granted, free of charge, to any person obtaining a copy
 *  of this software and associated documentation files (the "Software"), to
 *  deal in the Software without restriction, including without limitation the
 *  rights to use, copy, modify, merge, publish, distribute, sublicense, and/or
 *  sell copies of the Software, and to permit persons to whom the Software is
 *  furnished to do so, subject to the following conditions:
 *
 *  The above copyright notice and this permission notice shall be included in all
 *  copies or substantial portions of the Software.
 *
 *  THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 *  IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 *  FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 *  AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 *  LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
 *  FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS
 *  IN THE SOFTWARE.
 *
 **********************************************************************************************************************/
/**
 ***********************************************************************************************************************
 * @file  ResourceUsage.h
 * @brief LLPC header file: contains declaration of ResourceUsage and InterfaceData structs
 ***********************************************************************************************************************
 */
#pragma once

#include "lgc/state/Defs.h"
#include "lgc/state/IntrinsDefs.h"
#include "lgc/util/Internal.h"
#include <map>
#include <unordered_map>
#include <unordered_set>

namespace lgc {

static const unsigned MaxColorTargets = 8;

/// Represents the base data type
enum class BasicType : unsigned {
  Unknown = 0, // Unknown
  Float,       // Float
  Double,      // Double
  Int,         // Signed integer
  Uint,        // Unsigned integer
  Int64,       // 64-bit signed integer
  Uint64,      // 64-bit unsigned integer
  Float16,     // 16-bit floating-point
  Int16,       // 16-bit signed integer
  Uint16,      // 16-bit unsigned integer
  Int8,        // 8-bit signed integer
  Uint8,       // 8-bit unsigned integer
};

// Represents descriptor set/binding pair.
union DescriptorPair {
  struct {
    unsigned descSet; // ID of descriptor set
    unsigned binding; // ID of descriptor binding
  };
  uint64_t u64All;
};

// Represents transform feedback output info
union XfbOutInfo {
  struct {
    unsigned streamId : 2;   // Output stream ID
    unsigned xfbBuffer : 2;  // Transform feedback buffer
    unsigned xfbOffset : 27; // Transform feedback offset
    unsigned is16bit : 1;    // Whether it is 16-bit data for transform feedback
  };
  unsigned u32All;
};

// Represents interpolation info of fragment shader input
struct FsInterpInfo {
  unsigned loc;        // Mapped input location (tightly packed)
  bool flat;           // Whether it is "flat" interpolation
  bool custom;         // Whether it is "custom" interpolation
  bool is16bit;        // Whether it is 16-bit interpolation
  bool attr0Valid;     // Whether the location has a valid low half
  bool attr1Valid;     // Wheterh the location has a valid high half
  bool isPerPrimitive; // Whether it is per-primitive
};

// Invalid interpolation info
static const FsInterpInfo InvalidFsInterpInfo = {InvalidValue, false, false, false, false, false, false};

// Represents the location information on an input or output
class InOutLocationInfo {
public:
  InOutLocationInfo() { m_data.u16All = 0; }
  InOutLocationInfo(unsigned data) { m_data.u16All = data; }
  InOutLocationInfo(const InOutLocationInfo &inOutLocInfo) { m_data.u16All = inOutLocInfo.getData(); }
  InOutLocationInfo &operator=(const InOutLocationInfo &InOutLocationInfo) {
    m_data.u16All = InOutLocationInfo.getData();
    return *this;
  }

  unsigned getData() const { return static_cast<uint16_t>(m_data.u16All); }
  void setData(unsigned data) { m_data.u16All = static_cast<uint16_t>(data); }
  bool isInvalid() const { return m_data.u16All == 0xFFFF; }

  bool isHighHalf() const { return m_data.bits.isHighHalf; }
  void setHighHalf(bool isHighHalf) { m_data.bits.isHighHalf = isHighHalf; }

  unsigned getComponent() const { return m_data.bits.component; }
  void setComponent(unsigned compIdx) {
    assert(compIdx < 4); // Valid component offsets are 0~3
    m_data.bits.component = static_cast<uint16_t>(compIdx);
  }

  unsigned getLocation() const { return m_data.bits.location; }
  void setLocation(unsigned loc) { m_data.bits.location = static_cast<uint16_t>(loc); }

  bool isBuiltIn() const { return m_data.bits.isBuiltIn; }
  void setBuiltIn(bool isBuiltIn) { m_data.bits.isBuiltIn = isBuiltIn; }

  unsigned getStreamId() const { return m_data.bits.streamId; }
  void setStreamId(unsigned streamId) { m_data.bits.streamId = static_cast<uint16_t>(streamId); }

  bool operator<(const InOutLocationInfo &rhs) const { return this->getData() < rhs.getData(); }
  bool operator!=(const InOutLocationInfo &rhs) const { return this->getData() != rhs.getData(); }

private:
  union {
    struct {
      uint16_t isHighHalf : 1; // High half in case of 16-bit attributes
      uint16_t component : 2;  // The component index
      uint16_t location : 10;  // The location
      uint16_t isBuiltIn : 1;  // Whether location is actually built-in ID
      uint16_t streamId : 2;   // Output vertex stream ID
    } bits;
    uint16_t u16All;
  } m_data;
};

// Enumerate the workgroup layout options.
enum class WorkgroupLayout : unsigned {
  Unknown = 0,   // ?x?
  Linear,        // 4x1
  Quads,         // 2x2
  SexagintiQuads // 8x8
};

// Enumerate the workgroup layout for macro-tiling and
// Workgroup layout for micro-tiling when swizzle thread id.
struct SwizzleWorkgroupLayout {
  WorkgroupLayout microLayout;
  WorkgroupLayout macroLayout;
};

// Represents the usage info of shader resources.
//
// NOTE: All fields must be initialized in InitShaderResourceUsage().
struct ResourceUsage {
  std::unordered_set<uint64_t> descPairs;  // Pairs of descriptor set/binding
  bool resourceWrite = false;              // Whether shader does resource-write operations (UAV)
  bool resourceRead = false;               // Whether shader does resource-read operations (UAV)
  unsigned numSgprsAvailable = UINT32_MAX; // Number of available SGPRs
  unsigned numVgprsAvailable = UINT32_MAX; // Number of available VGPRs
  bool useImages = false;                  // Whether images are used
  bool useImageOp = false;                 // Whether image instruction is called (for GFX11+, pixel wait sync+)
  bool useRayQueryLdsStack = false;        // Whether ray query uses LDS stack

  // Usage of built-ins
  struct {
    // Per-stage built-in usage
    union {
      // Task shader
      struct {
        // Statement
        unsigned meshLinearDispatch : 1; // Mesh linear dispatch from task shader when group count Y/Z are both ones
        // Workgroup layout
        unsigned foldWorkgroupXY : 1; // The layout of the workgroup
      } task;

      // Vertex shader
      struct {
        // Input
        unsigned vertexIndex : 1;   // Whether gl_VertexIndex is used
        unsigned instanceIndex : 1; // Whether gl_InstanceIndex is used
        unsigned baseVertex : 1;    // Whether gl_BaseVertex is used
        unsigned baseInstance : 1;  // Whether gl_BaseInstance is used
        unsigned primitiveId : 1;   // Whether an implicit gl_PrimitiveID is required
        unsigned viewIndex : 1;     // Whether gl_ViewIndex is used
        // Output
        unsigned pointSize : 1;            // Whether gl_PointSize is used
        unsigned position : 1;             // Whether gl_Position is used
        unsigned clipDistance : 4;         // Array size of gl_ClipDistance[] (0 means unused)
        unsigned cullDistance : 4;         // Array size of gl_CullDistance[] (0 means unused)
        unsigned viewportIndex : 1;        // Whether gl_ViewportIndex is used
        unsigned layer : 1;                // Whether gl_Layer is used
        unsigned primitiveShadingRate : 1; // Whether gl_PrimitiveShadingRate is used
        unsigned edgeFlag : 1;             // Whether EdgeFlag is used
      } vs;

      // Tessellation control shader
      struct {
        // Input
        unsigned pointSizeIn : 1;     // Whether gl_in[].gl_PointSize is used
        unsigned positionIn : 1;      // Whether gl_in[].gl_Position is used
        unsigned clipDistanceIn : 4;  // Array size of gl_in[].gl_ClipDistance[] (0 means unused)
        unsigned cullDistanceIn : 4;  // Array size of gl_in[].gl_CullDistance[] (0 means unused)
        unsigned patchVertices : 1;   // Whether gl_PatchVerticesIn is used
        unsigned primitiveId : 1;     // Whether gl_PrimitiveID is used
        unsigned invocationId : 1;    // Whether gl_InvocationID is used
        unsigned viewIndex : 1;       // Whether gl_ViewIndex is used
        unsigned viewportIndexIn : 1; // Whether DX SV_ViewportArrayIndex is used
        unsigned layerIn : 1;         // Whether DX SV_RenderTargetArrayIndex is used
        // Output
        unsigned pointSize : 1;      // Whether gl_out[].gl_PointSize is used
        unsigned position : 1;       // Whether gl_out[].gl_Position is used
        unsigned clipDistance : 4;   // Array size of gl_out[].gl_ClipDistance[] (0 means unused)
        unsigned cullDistance : 4;   // Array size of gl_out[].gl_CullDistance[] (0 means unused)
        unsigned tessLevelOuter : 1; // Whether gl_TessLevelOuter[] is used
        unsigned tessLevelInner : 1; // Whether gl_TessLevelInner[] is used
        unsigned viewportIndex : 1;  // Whether DX SV_ViewportArrayIndex is used
        unsigned layer;              // Whether DX SV_RenderTargetArrayIndex is used
      } tcs;

      // Tessellation evaluation shader
      struct {
        // Input
        unsigned pointSizeIn : 1;     // Whether gl_in[].gl_PointSize is used
        unsigned positionIn : 1;      // Whether gl_in[].gl_Position is used
        unsigned clipDistanceIn : 4;  // Array size of gl_in[].gl_ClipDistance[] (0 means unused)
        unsigned cullDistanceIn : 4;  // Array size of gl_in[].gl_CullDistance[] (0 means unused)
        unsigned patchVertices : 1;   // Whether gl_PatchVerticesIn is used
        unsigned primitiveId : 1;     // Whether gl_PrimitiveID is used
        unsigned tessCoord : 1;       // Whether gl_TessCoord is used
        unsigned tessLevelOuter : 1;  // Whether gl_TessLevelOuter[] is used
        unsigned tessLevelInner : 1;  // Whether gl_TessLevelInner[] is used
        unsigned viewIndex : 1;       // Whether gl_ViewIndex is used
        unsigned viewportIndexIn : 1; // Whether DX SV_ViewportArrayIndex is used
        unsigned layerIn : 1;         // Whether DX SV_RenderTargetArrayIndex is used
        // Output
        unsigned pointSize : 1;     // Whether gl_PointSize is used
        unsigned position : 1;      // Whether gl_Position is used
        unsigned clipDistance : 4;  // Array size gl_ClipDistance[] (0 means unused)
        unsigned cullDistance : 4;  // Array size gl_CullDistance[] (0 means unused)
        unsigned viewportIndex : 1; // Whether gl_ViewportIndex is used
        unsigned layer : 1;         // Whether gl_Layer is used
      } tes;

      // Geometry shader
      struct {
        // Input
        unsigned pointSizeIn : 1;     // Whether gl_in[].gl_PointSize is used
        unsigned positionIn : 1;      // Whether gl_in[].gl_Position is used
        unsigned clipDistanceIn : 4;  // Array size of gl_in[].gl_ClipDistance[] (0 means unused)
        unsigned cullDistanceIn : 4;  // Array size of gl_in[].gl_CullDistance[] (0 means unused)
        unsigned primitiveIdIn : 1;   // Whether gl_PrimitiveIDIn is used
        unsigned invocationId : 1;    // Whether gl_InvocationID is used
        unsigned viewIndex : 1;       // Whether gl_ViewIndex is used
        unsigned viewportIndexIn : 1; // Whether DX SV_ViewportArrayIndex is used
        unsigned layerIn : 1;         // Whether DX SV_RenderTargetArrayIndex is used
        // Output
        unsigned pointSize : 1;            // Whether gl_PointSize is used
        unsigned position : 1;             // Whether gl_Position is used
        unsigned clipDistance : 4;         // Array size gl_ClipDistance[] (0 means unused)
        unsigned cullDistance : 4;         // Array size gl_CullDistance[] (0 means unused)
        unsigned primitiveId : 1;          // Whether gl_PrimitiveID is used
        unsigned viewportIndex : 1;        // Whether gl_ViewportIndex is used
        unsigned layer : 1;                // Whether gl_Layer is used
        unsigned primitiveShadingRate : 1; // Whether gl_PrimitiveShadingRate is used
      } gs;

      // Mesh shader
      struct {
        // Input
        unsigned drawIndex : 1;            // Whether gl_DrawIDARB is used
        unsigned viewIndex : 1;            // Whether gl_ViewIndex is used
        unsigned numWorkgroups : 1;        // Whether gl_NumWorkGroups is used
        unsigned workgroupId : 1;          // Whether gl_WorkGroupID is used
        unsigned localInvocationId : 1;    // Whether gl_LocalInvocationID is used
        unsigned globalInvocationId : 1;   // Whether gl_GlobalInvocationID is used
        unsigned localInvocationIndex : 1; // Whether gl_LocalInvocationIndex is used
        unsigned subgroupId : 1;           // Whether gl_SubgroupID is used
        unsigned numSubgroups : 1;         // Whether gl_NumSubgroups is used
        unsigned foldWorkgroupXY : 1;      // Whether the workgroup is folded
        // Output
        unsigned pointSize : 1;            // Whether gl_PointSize is used
        unsigned position : 1;             // Whether gl_Position is used
        unsigned clipDistance : 4;         // Array size gl_ClipDistance[] (0 means unused)
        unsigned cullDistance : 4;         // Array size gl_CullDistance[] (0 means unused)
        unsigned primitiveId : 1;          // Whether gl_PrimitiveID is used
        unsigned viewportIndex : 1;        // Whether gl_ViewportIndex is used
        unsigned layer : 1;                // Whether gl_Layer is used
        unsigned cullPrimitive : 1;        // Whether gl_CullPrimitive is used
        unsigned primitiveShadingRate : 1; // Whether gl_PrimitiveShadingRate is used
      } mesh;

      // Fragment shader
      struct {
        // Interpolation
        unsigned smooth : 1;        // Whether "smooth" qualifier is used
        unsigned noperspective : 1; // Whether "noperspective" qualifier is used
        unsigned flat : 1;          // Whether "flat" qualifier is used
        unsigned centroid : 1;      // Whether "centroid" qualifier is used
        unsigned sample : 1;        // Whether "sample" qualifier is used
        unsigned center : 1;        // Whether location qualifiers are not used (default: "center")
        unsigned pullMode : 1;      // Whether pull mode interpolation is used
        unsigned custom : 1;        // Whether custom interpolation is used
        // Input
        unsigned fragCoord : 1;                // Whether gl_FragCoord is used
        unsigned fragCoordIsSample : 1;        // Whether gl_FragCoord is used with sample Interpolation
        unsigned frontFacing : 1;              // Whether gl_FrontFacing is used
        unsigned clipDistance : 4;             // Array size of gl_ClipDistance[] (0 means unused)
        unsigned cullDistance : 4;             // Array size of gl_CullDistance[] (0 means unused)
        unsigned pointCoord : 1;               // Whether gl_PointCoord is used
        unsigned primitiveId : 1;              // Whether gl_PrimitiveID is used
        unsigned sampleId : 1;                 // Whether gl_SampleID is used
        unsigned samplePosition : 1;           // Whether gl_SamplePosition is used
        unsigned sampleMaskIn : 1;             // Whether gl_SampleMaskIn[] is used
        unsigned layer : 1;                    // Whether gl_Layer is used
        unsigned viewportIndex : 1;            // Whether gl_ViewportIndex is used
        unsigned helperInvocation : 1;         // Whether gl_HelperInvocation is used
        unsigned viewIndex : 1;                // Whether gl_ViewIndex is used
        unsigned shadingRate : 1;              // Whether gl_ShadingRate is used
        unsigned baryCoordNoPersp : 1;         // Whether gl_BaryCoordNoPersp is used (AMD extension)
        unsigned baryCoordNoPerspCentroid : 1; // Whether gl_BaryCoordNoPerspCentroid is used (AMD extension)
        unsigned baryCoordNoPerspSample : 1;   // Whether gl_BaryCoordNoPerspSample is used (AMD extension)
        unsigned baryCoordSmooth : 1;          // Whether gl_BaryCoordSmooth is used (AMD extension)
        unsigned baryCoordSmoothCentroid : 1;  // Whether gl_BaryCoordSmoothCentroid is used (AMD extension)
        unsigned baryCoordSmoothSample : 1;    // Whether gl_BaryCoordSmoothSample is used (AMD extension)
        unsigned baryCoordPullModel : 1;       // Whether gl_BaryCoordPullModel is used (AMD extension)
        unsigned baryCoord : 1;                // Whether gl_BaryCoordKHR is used
        unsigned baryCoordNoPerspKHR : 1;      // Whether gl_BaryCoordNoPerspKHR is used, distinction from
                                               // gl_BaryCoordNoPersp
        unsigned primType : 1;                 // Whether primTy is used, for OGL special FS emulation usage.
        unsigned primCoord : 1;                // Whether primCoord is used, for OGL internal FS emulation usage.
        unsigned lineStipple : 1;              // Whether lineStipple is used, for OGL internal FS emulation usage.
        // Output
        unsigned fragDepth : 1;      // Whether gl_FragDepth is used
        unsigned sampleMask : 1;     // Whether gl_SampleMask[] is used
        unsigned fragStencilRef : 1; // Whether gl_FragStencilRef is used
        // Statements
        unsigned discard : 1;           // Whether "discard" statement is used
        unsigned runAtSampleRate : 1;   // Whether fragment shader run at sample rate
        unsigned useDynamicToplogy : 1; // Whether to use dynamic topology.
      } fs;

      // Compute shader
      struct {
        // Workgroup layout
        unsigned foldWorkgroupXY : 1; // The layout of the workgroup
      } cs;
    };

  } builtInUsage;

  // Usage of generic input/output
  struct {
    // Map from shader specified InOutLocations to tightly packed InOutLocations
    std::map<InOutLocationInfo, InOutLocationInfo> inputLocInfoMap;
    std::map<InOutLocationInfo, InOutLocationInfo> outputLocInfoMap;

    std::map<unsigned, unsigned> perPatchInputLocMap;
    std::map<unsigned, unsigned> perPatchOutputLocMap;

    std::map<unsigned, unsigned> perPrimitiveInputLocMap;
    std::map<unsigned, unsigned> perPrimitiveOutputLocMap;

    // Map from built-in IDs to specially assigned locations
    std::map<unsigned, unsigned> builtInInputLocMap;
    std::map<unsigned, unsigned> builtInOutputLocMap;

    std::map<unsigned, unsigned> perPatchBuiltInInputLocMap;
    std::map<unsigned, unsigned> perPatchBuiltInOutputLocMap;

    std::map<unsigned, unsigned> perPrimitiveBuiltInInputLocMap;
    std::map<unsigned, unsigned> perPrimitiveBuiltInOutputLocMap;

    // Map from output location info to the transform feedback info
    std::map<InOutLocationInfo, XfbOutInfo> locInfoXfbOutInfoMap;

    // Count of transform feedback export call (each call is to export <4 x dword> at most), used in SW emulated
    // stream-out for GFX11+
    unsigned xfbExpCount = 0;

    // Count of mapped location for inputs/outputs (including those special locations to which the built-ins
    // are mapped)
    unsigned inputMapLocCount = 0;
    unsigned outputMapLocCount = 0;
    unsigned perPatchInputMapLocCount = 0;
    unsigned perPatchOutputMapLocCount = 0;
    unsigned perPrimitiveInputMapLocCount = 0;
    unsigned perPrimitiveOutputMapLocCount = 0;

    unsigned expCount = 0;     // Export count (number of "exp" instructions) for generic per-vertex outputs
    unsigned primExpCount = 0; // Export count (number of "exp" instructions) for generic per-primitive outputs

    struct {
      struct {
        // On-chip configurations (in dwords)
        struct {
          unsigned hsPatchCountStart; // Offset into LDS where HS patch count starts

          unsigned specialTfValueStart; // Offset into LDS where special TF values start
          unsigned specialTfValueSize;  // Size of special TF value (in dword)

          unsigned tessFactorStart;  // Offset into LDS where TFs start
          unsigned tessFactorStride; // Size of tess factor stride (in dword)

          unsigned outputPatchStart;   // Offset into LDS where vertices of output patches start
          unsigned outputVertexStride; // Stride of vertices of output patch (in dwords)
          unsigned outputPatchSize;    // Size of an output patch output (in dwords)

          unsigned patchConstStart; // Offset into LDS where patch constants start
          unsigned patchConstSize;  // Size of an output patch constant (in dwords)

          unsigned inputPatchStart;   // Offset into LDS where vertices of input patches start
          unsigned inputVertexStride; // Stride of vertices of input patch (in dwords)
          unsigned inputPatchSize;    // Size of an input patch size (in dwords)
        } onChip;

        // Off-chip configurations (in dwords)
        struct {
          unsigned outputPatchStart;   // Offset into LDS where vertices of output patches start
          unsigned outputVertexStride; // Stride of vertices of output patch (in dwords)
          unsigned outputPatchSize;    // Size of an output patch output (in dwords)

          unsigned patchConstStart; // Offset into LDS where patch constants start
          unsigned patchConstSize;  // Size of an output patch constant (in dwords)
        } offChip;

        unsigned maxNumPatchesPerGroup; // Maximum number of patches per thread group

        unsigned tessOnChipLdsSize;    // On-chip LDS size (exclude off-chip LDS buffer) (in dword)
        unsigned rayQueryLdsStackSize; // Ray query LDS stack size

        bool initialized; // Whether hwConfig has been initialized
      } hwConfig;
    } tcs = {};

    struct {
      // Map from IDs of built-in outputs to locations of generic outputs (used by copy shader to export built-in
      // outputs to fragment shader, always from vertex stream 0)
      std::map<unsigned, unsigned> builtInOutLocs;

      // Map from tightly packed locations to byte sizes of generic outputs (used by copy shader to
      // export generic outputs to fragment shader, always from vertex stream 0):
      //   <location, <component, byteSize>>
      std::unordered_map<unsigned, std::vector<unsigned>> genericOutByteSizes[MaxGsStreams];

      struct {
        unsigned esGsRingItemSize;                 // Size of each vertex written to the ES -> GS Ring, in dwords.
        unsigned gsVsRingItemSize;                 // Size of each primitive written to the GS -> VS Ring, in dwords.
        unsigned gsVsVertexItemSize[MaxGsStreams]; // Size of vertex item in the GS -> VS Ring, in dwords.
        unsigned esVertsPerSubgroup;               // Number of vertices ES exports.
        unsigned gsPrimsPerSubgroup;               // Number of prims GS exports.
        unsigned esGsLdsSize;                      // ES -> GS ring LDS size (GS in)
        unsigned gsOnChipLdsSize;                  // Total LDS size for GS on-chip mode.
        unsigned inputVertices;                    // Number of GS input vertices
        unsigned primAmpFactor;                    // GS primitive amplification factor
        bool enableMaxVertOut;                     // Whether to allow each GS instance to emit maximum vertices (NGG)
        unsigned rayQueryLdsStackSize;             // Ray query LDS stack size
      } hwConfig = {};

      unsigned outLocCount[MaxGsStreams] = {};
    } gs;

    struct {
      // Map from output locations to their number of components: <location, <numComponents, forBuiltIn>> (including
      // those special outputs to which built-ins are mapped)
      std::map<unsigned, std::pair<unsigned, BuiltInKind>> vertexOutputComponents;
      std::map<unsigned, std::pair<unsigned, BuiltInKind>> primitiveOutputComponents;
    } mesh;

    struct {
      std::vector<FsInterpInfo> interpInfo;   // Array of interpolation info
      BasicType outputTypes[MaxColorTargets]; // Array of basic types of fragment outputs
      bool isNullFs;                          // Is null FS, so should set final cbShaderMask to 0
    } fs;
  } inOutUsage;

  ResourceUsage(ShaderStageEnum shaderStage);
};

// Represents stream-out data
struct StreamOutData {
  unsigned tablePtr;                                   // Table pointer for stream-out
  unsigned controlBufPtr;                              // Control buffer pointer for stream-out (GFX11+)
  unsigned streamInfo;                                 // Stream-out info (ID, vertex count, enablement)
  unsigned writeIndex;                                 // Write index for stream-out
  unsigned streamOffsets[MaxTransformFeedbackBuffers]; // Stream-out Offset
};

// Represents interface data used by shader stages
//
// NOTE: All fields must be initialized in InitShaderInterfaceData().
struct InterfaceData {
  static const unsigned MaxDescTableCount = 64; // Must greater than (vk::MaxDynamicDescriptors +
                                                // vk::MaxDescriptorSets + special descriptors)
  static const unsigned MaxUserDataCount = 32;  // Max count of allowed user data (consider GFX IP version info)
  static const unsigned MaxSpillTableSize = 512;
  static const unsigned MaxDynDescCount = 32;
  static const unsigned MaxEsGsOffsetCount = 6;
  static const unsigned MaxCsUserDataCount = 16;

  unsigned userDataCount = 0; // User data count

  struct {
    unsigned sizeInDwords = 0; // Spill table size in dwords
  } spillTable;

  // Usage of user data registers for internal-use variables
  struct {
    // Geometry shader
    struct {
      unsigned copyShaderStreamOutTable; // Stream-out table (for copy shader)
    } gs;

    unsigned spillTable; // Spill table user data map

  } userDataUsage = {};

  // Indices of the arguments in shader entry-point
  struct {
    union {
      // Task shader
      struct {
        unsigned dispatchDims;       // Dispatch dimensions
        unsigned baseRingEntryIndex; // Base entry index (first workgroup) of mesh/task shader ring for current dispatch
        unsigned pipeStatsBuf;       // Pipeline statistics buffer
        unsigned workgroupId;        // Workgroup ID (GFX11 and below)
        unsigned multiDispatchInfo;  // Multiple dispatch info
        unsigned localInvocationId;  // Local invocation ID
      } task;

      // Vertex shader
      struct {
        unsigned baseVertex;         // Base vertex
        unsigned baseInstance;       // Base instance
        unsigned vertexId;           // Vertex ID
        unsigned relVertexId;        // Relative vertex ID (index of vertex within thread group)
        unsigned instanceId;         // Instance ID
        unsigned primitiveId;        // Primitive ID
        unsigned viewId;             // View ID
        unsigned vbTablePtr;         // Pointer of vertex buffer table
        unsigned esGsOffset;         // ES-GS ring buffer offset
        unsigned compositeData;      // CompositeData
        StreamOutData streamOutData; // Stream-out Data
      } vs;

      // Tessellation control shader
      struct {
        unsigned patchId;        // Patch ID
        unsigned relPatchId;     // Relative patch ID (control point ID included)
        unsigned tfBufferBase;   // Base offset of tessellation factor(TF) buffer
        unsigned offChipLdsBase; // Base offset of off-chip LDS buffer
        unsigned viewId;         // View ID
      } tcs;

      // Tessellation evaluation shader
      struct {
        unsigned tessCoordX;         // X channel of gl_TessCoord (U)
        unsigned tessCoordY;         // Y channel of gl_TessCoord (V)
        unsigned relPatchId;         // Relative patch id
        unsigned patchId;            // Patch ID
        unsigned esGsOffset;         // ES-GS ring buffer offset
        unsigned offChipLdsBase;     // Base offset of off-chip LDS buffer
        unsigned viewId;             // View ID
        StreamOutData streamOutData; // Stream-out Data
      } tes;

      // Geometry shader
      struct {
        unsigned gsVsOffset;                      // GS -> VS ring offset
        unsigned gsWaveId;                        // GS wave ID
        unsigned esGsOffsets[MaxEsGsOffsetCount]; // ES -> GS ring offset
        unsigned primitiveId;                     // Primitive ID
        unsigned invocationId;                    // Invocation ID
        unsigned viewId;                          // View ID
        unsigned compositeData;                   // CompositeData
        StreamOutData streamOutData;              // Stream-out Data
      } gs;

      // Mesh shader
      struct {
        unsigned drawIndex;          // Draw index
        unsigned viewId;             // View ID
        unsigned dispatchDims;       // Dispatch dimensions
        unsigned baseRingEntryIndex; // Base entry index (first workgroup) of mesh/task shader ring for current dispatch
        unsigned pipeStatsBuf;       // Pipeline statistics buffer
        unsigned flatWorkgroupId;    // Flat workgroup ID (emulated by HW vertex ID)
        unsigned localInvocationId;  // Local invocation ID (GFX11+, packed)
      } mesh;

      // Fragment shader
      struct {
        unsigned viewId;          // View ID
        unsigned primMask;        // Primitive mask
        unsigned collisionWaveId; // POPS collision wave ID (pre-GFX11)
        unsigned compositeData;   // CompositeData

        // Perspective interpolation (I/J)
        struct {
          unsigned sample;   // Sample
          unsigned center;   // Center
          unsigned centroid; // Centroid
          unsigned pullMode; // Pull-mode
        } perspInterp;

        // Linear interpolation (I/J)
        struct {
          unsigned sample;   // Sample
          unsigned center;   // Center
          unsigned centroid; // Centroid
        } linearInterp;

        unsigned lineStipple; // Line Stipple Tex Ena (f32)

        // FragCoord
        struct {
          unsigned x; // X channel
          unsigned y; // Y channel
          unsigned z; // Z channel
          unsigned w; // W channel
        } fragCoord;

        unsigned frontFacing;      // FrontFacing
        unsigned ancillary;        // Ancillary
        unsigned sampleCoverage;   // Sample coverage
        unsigned provokingVtxInfo; // Provoking vertex info
      } fs;

      // Compute shader
      struct {
        unsigned workgroupId;       // Workgroup ID
        unsigned multiDispatchInfo; // Multiple dispatch info
        unsigned localInvocationId; // Local invocation ID
      } cs;
    };

    bool initialized; // Whether entryArgIdxs has been initialized
                      //   by MutateEntryPoint
  } entryArgIdxs = {};

  InterfaceData();
};

} // namespace lgc
