/*
 ***********************************************************************************************************************
 *
 *  Copyright (c) 2018-2025 Advanced Micro Devices, Inc. All Rights Reserved.
 *
 *  Permission is hereby granted, free of charge, to any person obtaining a copy
 *  of this software and associated documentation files (the "Software"), to
 *  deal in the Software without restriction, including without limitation the
 *  rights to use, copy, modify, merge, publish, distribute, sublicense, and/or
 *  sell copies of the Software, and to permit persons to whom the Software is
 *  furnished to do so, subject to the following conditions:
 *
 *  The above copyright notice and this permission notice shall be included in all
 *  copies or substantial portions of the Software.
 *
 *  THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 *  IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 *  FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 *  AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 *  LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
 *  FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS
 *  IN THE SOFTWARE.
 *
 **********************************************************************************************************************/
/**
 ***********************************************************************************************************************
 * @file  NggPrimShader.h
 * @brief LLPC header file: contains declaration of class lgc::NggPrimShader.
 ***********************************************************************************************************************
 */
#pragma once

#include "lgc/state/PipelineState.h"
#include "lgc/state/TargetInfo.h"
#include "lgc/util/BuilderBase.h"
#include "llvm/IR/Module.h"

namespace lgc {

struct NggControl;
class PipelineState;

// Represents constant buffer offsets (in bytes) of viewport controls in primitive shader table.
// NOTE: The layout structure is defined by @ref Util::Abi::PrimShaderVportCb.
struct PrimShaderVportCbLookupTable {
  // Viewport transform scale
  unsigned paClVportXscale;
  unsigned paClVportXoffset;
  unsigned paClVportYscale;
  unsigned paClVportYoffset;
  // Viewport width/height
  unsigned vportWidth;
  unsigned vportHeight;
};

// Represents a collection of constant buffer offsets (in bytes) in primitive shader table.
// NOTE: The layout structure is defined by @ref Util::Abi::PrimShaderCbLayout.
struct PrimShaderCbLayoutLookupTable {
  // GS addressed used for jump from ES
  unsigned gsAddressLo;
  unsigned gsAddressHi;
  // Viewport transform controls
  unsigned paClVteCntl;
  // Float-to-fixed-vertex conversion controls
  unsigned paSuVtxCntl;
  // Clip space controls
  unsigned paClClipCntl;
  // Culling controls
  unsigned paSuScModeCntl;
  // Various frustum culling controls
  unsigned paClGbHorzClipAdj; // Horizontal adjacent culling control
  unsigned paClGbVertClipAdj; // Vertical adjacent culling control
  unsigned paClGbHorzDiscAdj; // Horizontal discard culling control
  unsigned paClGbVertDiscAdj; // Vertical discard culling control
  // Run-time handling primitive type
  unsigned vgtPrimitiveType;
  // Number of MSAA samples
  unsigned msaaNumSamples;
  // Render state
  unsigned primitiveRestartEnable;
  unsigned primitiveRestartIndex;
  unsigned matchAllBits;
  unsigned enableConservativeRasterization;
  // Viewport controls
  PrimShaderVportCbLookupTable vportControls[Util::Abi::MaxViewports];
};

// Represents the layout structure of an item of vertex cull info (this acts as ES-GS ring item from HW's perspective)
struct VertexCullInfo {
  //
  // Vertex transform feedback outputs
  //
  unsigned xfbOutputs[4]; // At most <4 x dword>
  //
  // Vertex cull data
  //
  unsigned cullDistanceSignMask;
  //
  // Vertex cull result
  //
  unsigned drawFlag;
  //
  // Vertex compaction info (vertex compaction only, must in the end of this structure)
  //
  unsigned compactedVertexIndex;
  union {
    struct {
      unsigned vertexId;
      unsigned instanceId;
      unsigned primitiveId;
    } vs;
    struct {
      float tessCoordX;
      float tessCoordY;
      unsigned patchId;
      unsigned relPatchId;
    } tes;
  };
};

// Represents a collection of LDS offsets (in dwords) within an item of vertex cull info.
struct VertexCullInfoOffsets {
  //
  // Vertex transform feedback outputs
  //
  unsigned xfbOutputs;
  //
  // Vertex cull data
  //
  unsigned cullDistanceSignMask;
  //
  // Vertex cull result
  //
  unsigned drawFlag;
  //
  // Vertex compaction info
  //
  unsigned compactedVertexIndex;
  // VS
  unsigned vertexId;
  unsigned instanceId;
  unsigned primitiveId;
  // TES
  unsigned tessCoordX;
  unsigned tessCoordY;
  unsigned patchId;
  unsigned relPatchId;
};

// Represents export info of a vertex
struct VertexExport {
  unsigned exportSlot;      // Export slot
  unsigned channelMask;     // Channel mask
  llvm::Value *exportValue; // Export values
};

// Represents export info of a XFB output
struct XfbExport {
  unsigned xfbBuffer;   // XFB buffer
  unsigned xfbOffset;   // XFB offset
  unsigned numElements; // Number of export elements, valid range is [1,4]
  // For ES only
  unsigned offsetInVertex; // Offset of a XFB output within all XFB outputs of a vertex
  // For GS only
  struct {
    unsigned streamId;      // Output stream ID
    unsigned location;      // Output location
    unsigned component;     // Output component within a location
  } locInfo;                // Output location info in GS-VS ring
  llvm::Value *exportValue; // Export value of a XFB output
};

// Enumerates the LDS regions used by primitive shader
enum class PrimShaderLdsRegion : unsigned {
  DistributedPrimitiveId, // Distributed primitive ID
  XfbOutput,              // Transform feedback outputs
  VertexPosition,         // Vertex position
  VertexCullInfo,         // Vertex cull info
  XfbStats,               // Transform feedback statistics
  VertexCounts,           // Vertex counts in waves and in NGG subgroup
  VertexIndexMap,         // Vertex index map (compacted -> uncompacted)
  EsGsRing,               // ES-GS ring
  PrimitiveData,          // Primitive connectivity data
  PrimitiveCounts,        // Primitive counts in waves and in NGG subgroup
  PrimitiveIndexMap,      // Primitive index map (compacted -> uncompacted)
  GsVsRing,               // GS-VS ring
};

// Represents LDS usage info of primitive shader
struct PrimShaderLdsUsageInfo {
  bool needsLds;           // Whether primitive shader needs LDS for operations
  unsigned esExtraLdsSize; // ES extra LDS size in dwords
  unsigned gsExtraLdsSize; // GS extra LDS size in dwords
};

// Map: LDS region -> <region Offset, region Size>
typedef std::unordered_map<PrimShaderLdsRegion, std::pair<unsigned, unsigned>> PrimShaderLdsLayout;

// Represents a collection of constant buffer offsets (in dwords) within stream-out control buffer.
// NOTE: The layout structure is defined by @ref Util::Abi::StreamOutControlCb.
struct StreamOutControlCbOffsets {
  unsigned bufOffsets[MaxTransformFeedbackBuffers];
#if LLPC_BUILD_GFX12
  // Following offsets are only available on GFX12+ (caused by GDS removal)
  unsigned primsNeeded[MaxGsStreams];
  unsigned primsWritten[MaxGsStreams];
  struct {
    unsigned orderedWaveId;
    unsigned dwordsWritten;
  } orderedIdPair[MaxTransformFeedbackBuffers];
#endif
};

// =====================================================================================================================
// Represents the manager of NGG primitive shader.
class NggPrimShader {
public:
  NggPrimShader(PipelineState *pipelineState);

  static unsigned calcEsGsRingItemSize(PipelineState *pipelineState, llvm::Function *esMain);
  static PrimShaderLdsUsageInfo layoutPrimShaderLds(PipelineState *pipelineState,
                                                    PrimShaderLdsLayout *ldsLayout = nullptr);

  llvm::Function *generate(llvm::Function *esMain, llvm::Function *gsMain, llvm::Function *copyShader);

private:
  NggPrimShader() = delete;
  NggPrimShader(const NggPrimShader &) = delete;
  NggPrimShader &operator=(const NggPrimShader &) = delete;

  static unsigned calcVertexCullInfoSizeAndOffsets(PipelineState *pipelineState, llvm::Function *esMain,
                                                   VertexCullInfoOffsets &vertCullInfoOffsets);
  static unsigned calcEsXfbOutputsSize(llvm::Function *target);

  llvm::FunctionType *getPrimShaderType(uint64_t &inRegMask);

  void buildPrimShaderCbLayoutLookupTable();
  void calcStreamOutControlCbOffsets();

  void buildPassthroughPrimShader(llvm::Function *entryPoint);
  void buildPrimShader(llvm::Function *entryPoint);
  void buildPrimShaderWithGs(llvm::Function *entryPoint);

  void initWaveThreadInfo(llvm::Value *mergedGroupInfo, llvm::Value *mergedWaveInfo);
  void prepareAttribRingAccess(llvm::Value *userData);
  void loadStreamOutBufferInfo(llvm::Value *userData);
  void distributePrimitiveId(llvm::Value *primitiveId);

  llvm::Value *cullPrimitive(llvm::Value *vertexIndex0, llvm::Value *vertexIndex1, llvm::Value *vertexIndex2);
  void sendGsAllocReqMessage();
  void exportPassthroughPrimitive();
  void exportPrimitive(llvm::Value *primitiveCulled);
  void exportPrimitiveWithGs(llvm::Value *startingVertexIndex);

  void earlyExitWithDummyExport();

  void runEs(llvm::ArrayRef<llvm::Argument *> args);
  void splitEs();
  void runGs(llvm::ArrayRef<llvm::Argument *> args);
  void mutateGs();
  void mutateCopyShader();

  void appendUserData(llvm::SmallVectorImpl<llvm::Value *> &args, llvm::Function *target, llvm::Value *userData,
                      unsigned userDataCount);

  void writeGsOutput(llvm::Value *output, unsigned location, unsigned component, unsigned streamId,
                     llvm::Value *primitiveIndex, llvm::Value *emitVerts, llvm::Value *totalEmitVertsPtr);
  llvm::Value *readGsOutput(llvm::Type *outputTy, unsigned location, unsigned component, unsigned streamId,
                            llvm::Value *vertexOffset);

  void processGsEmit(unsigned streamId, llvm::Value *primitiveIndex, llvm::Value *emitVertsPtr,
                     llvm::Value *outVertsPtr, llvm::Value *totalEmitVertsPtr);
  void processGsCut(unsigned streamId, llvm::Value *outVertsPtr);

  llvm::Function *createGsEmitHandler();
  llvm::Function *createGsCutHandler();

  llvm::Value *readPerThreadDataFromLds(llvm::Type *readDataTy, llvm::Value *threadId, PrimShaderLdsRegion region,
                                        unsigned offsetInRegion = 0, bool useDs128 = false);
  void writePerThreadDataToLds(llvm::Value *writeData, llvm::Value *threadId, PrimShaderLdsRegion region,
                               unsigned offsetInRegion = 0, bool useDs128 = false);

  llvm::Value *readVertexCullInfoFromLds(llvm::Type *readDataTy, llvm::Value *vertexItemOffset, unsigned dataOffset);
  void writeVertexCullInfoToLds(llvm::Value *writeData, llvm::Value *vertexItemOffset, unsigned dataOffset);

  llvm::Value *runBackfaceCuller(llvm::Value *primitiveAlreadyCulled, llvm::Value *vertex0, llvm::Value *vertex1,
                                 llvm::Value *vertex2);
  llvm::Value *runFrustumCuller(llvm::Value *primitiveAlreadyCulled, llvm::Value *vertex0, llvm::Value *vertex1,
                                llvm::Value *vertex2);
  llvm::Value *runBoxFilterCuller(llvm::Value *primitiveAlreadyCulled, llvm::Value *vertex0, llvm::Value *vertex1,
                                  llvm::Value *vertex2);
  llvm::Value *runSphereCuller(llvm::Value *primitiveAlreadyCulled, llvm::Value *vertex0, llvm::Value *vertex1,
                               llvm::Value *vertex2);
  llvm::Value *runSmallPrimFilterCuller(llvm::Value *primitiveAlreadyCulled, llvm::Value *vertex0, llvm::Value *vertex1,
                                        llvm::Value *vertex2);
  llvm::Value *runCullDistanceCuller(llvm::Value *primitiveAlreadyCulled, llvm::Value *signMask0,
                                     llvm::Value *signMask1, llvm::Value *signMask2);
  llvm::Value *fetchCullingControlRegister(unsigned regOffset);

  llvm::Function *createBackfaceCuller();
  llvm::Function *createFrustumCuller();
  llvm::Function *createBoxFilterCuller();
  llvm::Function *createSphereCuller();
  llvm::Function *createSmallPrimFilterCuller();
  llvm::Function *createCullDistanceCuller();
  llvm::Function *createFetchCullingRegister();

  llvm::Value *ballot(llvm::Value *value);

  llvm::Value *fetchVertexPositionData(llvm::Value *vertexIndex);
  llvm::Value *fetchCullDistanceSignMask(llvm::Value *vertexIndex);
  llvm::Value *calcVertexItemOffset(unsigned streamId, llvm::Value *vertexIndex);

  llvm::Function *makeExportCollector(llvm::Function *&fromFunc, bool makeClone,
                                      llvm::SmallVectorImpl<VertexExport> *positionExports,
                                      llvm::SmallVectorImpl<VertexExport> *attributeExports,
                                      llvm::SmallVectorImpl<XfbExport> *xfbExports);
  void collectExports(llvm::ArrayRef<llvm::Argument *> args, llvm::Function *&fromFunc, bool makeClone,
                      llvm::SmallVectorImpl<VertexExport> *positionExports,
                      llvm::SmallVectorImpl<VertexExport> *attributeExports,
                      llvm::SmallVectorImpl<XfbExport> *xfbExports);
  void createPhiForExports(llvm::SmallVectorImpl<VertexExport> *positionExports,
                           llvm::SmallVectorImpl<VertexExport> *attributeExports,
                           llvm::SmallVectorImpl<XfbExport> *xfbExports);
  void exportPositions(const llvm::SmallVectorImpl<VertexExport> &positionExports);
  void exportAttributes(const llvm::SmallVectorImpl<VertexExport> &attributeExports);

  void processSwXfb(llvm::ArrayRef<llvm::Argument *> args, const llvm::SmallVectorImpl<XfbExport> &xfbExports);
  void processSwXfbWithGs(llvm::ArrayRef<llvm::Argument *> args, const llvm::SmallVectorImpl<XfbExport> &xfbExports);
  void prepareSwXfb(llvm::ArrayRef<llvm::Value *> primCountInSubgroup);

  llvm::Value *readXfbOutputFromLds(llvm::Type *readDataTy, llvm::Value *vertexIndex, unsigned offsetInVertex);
  void writeXfbOutputToLds(llvm::Value *writeData, llvm::Value *vertexIndex, unsigned offsetInVertex);

  // Checks if NGG culling operations are enabled
  bool enableCulling() const {
    return m_nggControl->enableBackfaceCulling || m_nggControl->enableFrustumCulling ||
           m_nggControl->enableBoxFilterCulling || m_nggControl->enableSphereCulling ||
           m_nggControl->enableSmallPrimFilter || m_nggControl->enableCullDistanceCulling;
  }

  llvm::BasicBlock *createBlock(llvm::Function *parent, const llvm::Twine &blockName = "");
  llvm::Value *createUBfe(llvm::Value *value, unsigned offset, unsigned count);
  llvm::Value *makePointer(llvm::Value *ptrValue, llvm::Type *ptrTy);
  llvm::PHINode *createPhi(llvm::ArrayRef<std::pair<llvm::Value *, llvm::BasicBlock *>> incomings,
                           const llvm::Twine &name = "");
  void createFenceAndBarrier();
  void createBarrier();

  unsigned getLdsRegionStart(PrimShaderLdsRegion region) {
    assert(m_ldsLayout.count(region) > 0);
    return m_ldsLayout[region].first;
  }

  void collectPrimitiveStats();

  llvm::Value *readValueFromLds(llvm::Type *readTy, llvm::Value *ldsOffset, bool useDs128 = false);
  void writeValueToLds(llvm::Value *writeValue, llvm::Value *ldsOffset, bool useDs128 = false);
  void atomicOp(llvm::AtomicRMWInst::BinOp atomicOp, llvm::Value *value, llvm::Value *ldsOffset);
  llvm::Value *readValueFromCb(llvm::Type *readyTy, llvm::Value *bufPtr, llvm::Value *offset, bool isVolatile = false);
#if LLPC_BUILD_GFX12
  llvm::Value *globalAtomicOp(llvm::AtomicRMWInst::BinOp atomicOp, llvm::Value *value, llvm::Value *basePtr,
                              llvm::Value *offset);
  std::pair<llvm::Value *, llvm::Value *> globalAtomicOrderedAdd(std::pair<llvm::Value *, llvm::Value *> orderedIdPair,
                                                                 llvm::Value *basePtr, llvm::Value *offset);
#endif

  static const unsigned NullPrim = (1u << 31); // Null primitive data (invalid)

  PipelineState *m_pipelineState = nullptr; // Pipeline state
  GfxIpVersion m_gfxIp;                     // Graphics IP version info

  const NggControl *m_nggControl = nullptr; // NGG control settings

  // NGG inputs (from system values or derived from them)
  struct {
    // SGPRs
    llvm::Value *vertCountInSubgroup; // Number of vertices in subgroup
    llvm::Value *primCountInSubgroup; // Number of primitives in subgroup
    llvm::Value *vertCountInWave;     // Number of vertices in wave
    llvm::Value *primCountInWave;     // Number of primitives in wave

    llvm::Value *waveIdInSubgroup; // Wave ID in subgroup
    llvm::Value *orderedWaveId;    // Ordered wave ID

    std::pair<llvm::Value *, llvm::Value *> primShaderTableAddr; // Primitive shader table address <low, high>

    // VGPRs
    llvm::Value *threadIdInWave;     // Thread ID in wave
    llvm::Value *threadIdInSubgroup; // Thread ID in subgroup

    llvm::Value *primData; // Primitive connectivity data (provided by HW)

    llvm::Value *vertexIndex0; // Relative index of vertex0 in NGG subgroup
    llvm::Value *vertexIndex1; // Relative index of vertex1 in NGG subgroup
    llvm::Value *vertexIndex2; // Relative index of vertex2 in NGG subgroup
  } m_nggInputs = {};

  // ES handlers
  struct {
    llvm::Function *main; // ES main function
    // Part ES functions, the first part contains cull data exports (position and cull distance) and the second part
    // contains other remaining exports.
    std::pair<llvm::Function *, llvm::Function *> part;
  } m_esHandlers = {};

  // GS handlers
  struct {
    llvm::Function *main;       // GS main function
    llvm::Function *copyShader; // Copy shader
    llvm::Function *emit;       // GS emit handler
    llvm::Function *cut;        // GS cut handler
  } m_gsHandlers = {};

  // Cullers
  struct {
    llvm::Function *backface;        // Backface culller
    llvm::Function *frustum;         // Frustum culler
    llvm::Function *boxFilter;       // Box filter culler
    llvm::Function *sphere;          // Sphere culler
    llvm::Function *smallPrimFilter; // Small primitive filter
    llvm::Function *cullDistance;    // Cull distance culler
    llvm::Function *regFetcher;      // Culling register fetcher
  } m_cullers = {};

  llvm::Value *m_distributedPrimitiveId = nullptr; // Distributed primitive ID (from geomeotry based to vertex based)

  llvm::Value *m_compactVertex = nullptr; // Flag indicating whether to perform vertex compaction (if
                                          // null, we are in vertex compactionless mode)

  bool m_hasVs = false;  // Whether the pipeline has vertex shader
  bool m_hasTes = false; // Whether the pipeline has tessellation evaluation shader
  bool m_hasGs = false;  // Whether the pipeline has geometry shader

  unsigned m_maxThreadsPerSubgroup = 0; // Maximum number of threads in a NGG subgroup
  unsigned m_maxWavesPerSubgroup = 0;   // Maximum number of waves in a NGG subgroup

  llvm::Value *m_attribRingBufDesc = nullptr;    // Attribute ring buffer descriptor
  llvm::Value *m_attribRingBaseOffset = nullptr; // Subgroup's attribute ring base offset (in bytes)

  llvm::Value *m_streamOutControlBufPtr = nullptr;                      // Stream-out control buffer pointer
  llvm::Value *m_streamOutBufDescs[MaxTransformFeedbackBuffers] = {};   // Stream-out buffer descriptors
  llvm::Value *m_streamOutBufOffsets[MaxTransformFeedbackBuffers] = {}; // Stream-out buffer offsets

  llvm::Value *m_verticesPerPrimitive = nullptr; // If topology is dynamic, it is a SGPR value from user data
                                                 // ComplexData; otherwise it is a constant.

  bool m_constPositionZ = false; // Whether the Z channel of vertex position data is constant

  // Base offsets (in dwords) of GS output vertex streams in GS-VS ring
  unsigned m_gsStreamBases[MaxGsStreams] = {};

  PrimShaderCbLayoutLookupTable m_cbLayoutTable;         // Layout lookup table of primitive shader constant buffer
  VertexCullInfoOffsets m_vertCullInfoOffsets;           // A collection of offsets within an item of vertex cull info
  StreamOutControlCbOffsets m_streamOutControlCbOffsets; // A collection of offsets within stream-out control buffer

  BuilderBase m_builder; // LLVM IR builder

  llvm::Constant *m_lds = nullptr; // Global variable to model primitive shader LDS
  PrimShaderLdsLayout m_ldsLayout; // Primitive shader LDS layout
};

} // namespace lgc
