//===- SPIRVReader.h - Converts SPIR-V to LLVM ----------------*- C++ -*-===//
//
//                     The LLVM/SPIR-V Translator
//
// This file is distributed under the University of Illinois Open Source
// License. See LICENSE.TXT for details.
//
// Copyright (c) 2014-2025 Advanced Micro Devices, Inc. All rights reserved.
//
// Permission is hereby granted, free of charge, to any person obtaining a
// copy of this software and associated documentation files (the "Software"),
// to deal with the Software without restriction, including without limitation
// the rights to use, copy, modify, merge, publish, distribute, sublicense,
// and/or sell copies of the Software, and to permit persons to whom the
// Software is furnished to do so, subject to the following conditions:
//
// Redistributions of source code must retain the above copyright notice,
// this list of conditions and the following disclaimers.
// Redistributions in binary form must reproduce the above copyright notice,
// this list of conditions and the following disclaimers in the documentation
// and/or other materials provided with the distribution.
// Neither the names of Advanced Micro Devices, Inc., nor the names of its
// contributors may be used to endorse or promote products derived from this
// Software without specific prior written permission.
// THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
// IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
// FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
// CONTRIBUTORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
// LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
// OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS WITH
// THE SOFTWARE.
//
//===----------------------------------------------------------------------===//
/// \file
///
/// This file implements conversion of SPIR-V binary to LLVM IR.
///
//===----------------------------------------------------------------------===//
#ifndef SPIRVREADER_H
#define SPIRVREADER_H

#include "SPIRVInstruction.h"
#include "SPIRVInternal.h"
#include "SPIRVModule.h"
#include "SPIRVToLLVMDbgTran.h"
#include "vkgcDefs.h"
#include "compilerutils/LoweringPointerTupleMap.h"
#include "lgc/Builder.h"
#include "lgc/LgcXdlTypes.h"

namespace llvm {
class Module;
class Type;
class Instruction;
class CallInst;
class BasicBlock;
class Function;
class GlobalVariable;
class LLVMContext;
class LoadInst;
class BranchInst;
class BinaryOperator;
class Value;
} // namespace llvm
using namespace llvm;

namespace Llpc {
class PipelineContext;
class RayTracingContext;
} // namespace Llpc

namespace SPIRV {
class SPIRVLoopMerge;
class SPIRVToLLVMDbgTran;

enum class LayoutMode : uint8_t {
  None = 0,     ///< SSA value -- has no memory layout
  Native = 1,   ///< Using native LLVM rules for in-memory layout
  Explicit = 2, ///< Using layout decorations(like offset) from SPIRV
  Std430 = 3,   ///< Using std430 layout rule
  Scalar = 4,   ///< Using scalar layout
};

// Describe what parts of image/sampler descriptors are present.
enum ImageComponent {
  ImageComponentImage = 0x1,
  ImageComponentFMask = 0x2,
  ImageComponentSampler = 0x4,
};

// Holds indices into a struct describing an image/sampler (pointer) value.
struct ImageTypeIndices {
  unsigned imagePointer = ~0;
  unsigned imageStride = ~0;
  unsigned imagePlaneStride = ~0;
  unsigned fmaskPointer = ~0;
  unsigned fmaskStride = ~0;
  unsigned samplerPointer = ~0;
  unsigned samplerStride = ~0;
  unsigned convertingSamplerIdx = ~0;
  unsigned compareParamPointer = ~0;
};

class SPIRVToLLVM {
public:
  SPIRVToLLVM(Module *llvmModule, SPIRVModule *theSpirvModule, const SPIRVSpecConstMap &theSpecConstMap,
              llvm::ArrayRef<ConvertingSampler> convertingSamplers, lgc::Builder *builder,
              const Vkgc::ShaderModuleUsage *moduleUsage, const Vkgc::PipelineShaderOptions *shaderOptions);

  void setGlobalVarPrefix(llvm::StringRef globalVarPrefix) { m_globalVarPrefix = globalVarPrefix; }

  DebugLoc getDebugLoc(SPIRVInstruction *bi, Function *f);

  void updateDebugLoc(SPIRVValue *bv, Function *f);

  unsigned getImageTypeComponents(SPIRVType *t) const;
  ImageTypeIndices getImageTypeIndices(unsigned imageComponents) const;
  Type *getImageTy(unsigned imageComponents) const;

  Type *transType(SPIRVType *bt, unsigned matrixStride = 0, bool columnMajor = true,
                  LayoutMode layout = LayoutMode::None);

  SmallVector<Type *> transMultiTypes(SPIRVType *bt, unsigned matrixStride = 0, bool columnMajor = true,
                                      LayoutMode layout = LayoutMode::None);

  bool mapMultiTypes(Type *pt, SPIRVType *t, unsigned matrixStride = 0, bool columnMajor = true,
                     LayoutMode layout = LayoutMode::None);

  template <spv::Op>
  Type *transTypeWithOpcode(SPIRVType *bt, unsigned matrixStride, bool columnMajor, LayoutMode layout);
  Type *transTypeArray(SPIRVType *bt, unsigned matrixStride, bool columnMajor, LayoutMode layout);
  std::vector<Type *> transTypeVector(const std::vector<SPIRVType *> &);
  bool translate(ExecutionModel entryExecModel, const char *entryName);
  bool transAddressingModel();

  SmallVector<Value *> transValueMulti(SPIRVValue *, Function *f, BasicBlock *bb, bool createPlaceHolder = true);
  Value *transValue(SPIRVValue *, Function *f, BasicBlock *, bool createPlaceHolder = true);
  SmallVector<Value *> transValueWithoutDecoration(SPIRVValue *, Function *f, BasicBlock *,
                                                   bool createPlaceHolder = true);
  Value *transAtomicRMW(SPIRVValue *, const AtomicRMWInst::BinOp);
  Constant *transInitializer(SPIRVValue *, Type *);
  template <spv::Op> Value *transValueWithOpcode(SPIRVValue *);
  template <spv::Op> Value *transValueWithOpcode(SPIRVValue *, Function *f, BasicBlock *bb);
  template <spv::Op> SmallVector<Value *> transValueMultiWithOpcode(SPIRVValue *);
  template <spv::Op> SmallVector<Value *> transValueMultiWithOpcode(SPIRVValue *, Function *f, BasicBlock *bb);
  Value *transLoadImage(SPIRVValue *spvImageLoadPtr);
  Value *transLoadBindlessImage(SPIRVType *spvElementTy, Value *imgDescGpuAddress, bool bindlessTexture);
  Value *transImagePointer(SPIRVValue *spvImagePtr, SPIRVType *elementTy = nullptr);
  Value *indexDescPtr(SPIRVType *spvElementTy, Value *base, Value *index);
  Value *transGroupArithOp(lgc::Builder::GroupArithOp, SPIRVValue *);

  bool transDecoration(SPIRVValue *, ArrayRef<Value *>);
  unsigned remapOutputLocation(unsigned loc);
  bool checkContains64BitType(SPIRVType *bt);
  Constant *buildShaderInOutMetadata(SPIRVType *bt, ShaderInOutDecorate &inOutDec, Type *&metaTy, bool vs64bitsAttrib);
  Constant *buildShaderBlockMetadata(SPIRVType *bt, ShaderBlockDecorate &blockDec, Type *&mdTy,
                                     SPIRVStorageClassKind storageClass);
  unsigned calcShaderBlockSize(SPIRVType *bt, unsigned blockSize, unsigned matrixStride, bool isRowMajor);
  Value *transGLSLExtInst(SPIRVExtInst *extInst, BasicBlock *bb);
  Value *flushDenorm(Value *val);
  Value *transTrinaryMinMaxExtInst(SPIRVExtInst *extInst, BasicBlock *bb);
  Value *transGLSLBuiltinFromExtInst(SPIRVExtInst *bc, BasicBlock *bb);
  std::vector<Value *> transValue(const std::vector<SPIRVValue *> &, Function *f, BasicBlock *);
  Function *transFunction(SPIRVFunction *f);
  bool transMetadata();
  bool transNonTemporalMetadata(Instruction *i);
  Value *transConvertInst(SPIRVValue *bv, Function *f, BasicBlock *bb);
  Instruction *transBuiltinFromInst(const std::string &funcName, SPIRVInstruction *bi, BasicBlock *bb);
  Instruction *transSPIRVBuiltinFromInst(SPIRVInstruction *bi, BasicBlock *bb);
  void transBarrierFence(SPIRVInstruction *bi, BasicBlock *bb);
  Value *transString(const SPIRVString *spvValue);
  Value *transDebugPrintf(SPIRVInstruction *bi, const ArrayRef<SPIRVValue *> spvValues, Function *func, BasicBlock *bb);
  Value *transVariableNonImage(SPIRVValue *const spvValue);
  SmallVector<Value *> transAccessChain(SPIRVValue *const spvValue);
  Value *transArrayLength(SPIRVValue *const spvValue);
  // Struct used to pass information in and out of getImageDesc.
  struct ExtractedImageInfo {
    const SPIRVTypeImageDescriptor *desc = nullptr;
    unsigned dim = ~0;  // lgc::Builder dimension
    unsigned flags = 0; // lgc::Builder image call flags
    Value *imagePointer = nullptr;
    Value *imagePlaneStride = nullptr;
    Value *fmaskPointer = nullptr;
    Value *samplerPointer = nullptr;
    Value *convertingSamplerIdx = nullptr;
    Value *compareParamPointer = nullptr;
  };

  // Load image and/or sampler descriptors, and get information from the image
  // type.
  void getImageDesc(SPIRVValue *bImageInst, ExtractedImageInfo *info);

  // Get number of channels based on the image format.
  // Returns 0 if number of channels is unknown.
  unsigned getImageNumChannels(const SPIRVTypeImageDescriptor *descriptor);

  // Set up address operand array for image sample/gather builder call.
  void setupImageAddressOperands(SPIRVInstruction *bi, unsigned maskIdx, bool hasProj, MutableArrayRef<Value *> addr,
                                 ExtractedImageInfo *imageInfo, Value **sampleNum);

  // Handle fetch/read/write/atomic aspects of coordinate.
  void handleImageFetchReadWriteCoord(SPIRVInstruction *bi, ExtractedImageInfo *imageInfo,
                                      MutableArrayRef<Value *> addr, bool enableMultiView = true);

  // Translate SPIR-V image atomic operations to LLVM function calls
  Value *transSPIRVImageAtomicOpFromInst(SPIRVInstruction *bi, BasicBlock *bb);

  // Translates SPIR-V fragment mask operations to LLVM function calls
  Value *transSPIRVFragmentMaskFetchFromInst(SPIRVInstruction *bi, BasicBlock *bb);
  Value *transSPIRVFragmentFetchFromInst(SPIRVInstruction *bi, BasicBlock *bb);

  // Translate image sample to LLVM IR
  Value *transSPIRVImageSampleFromInst(SPIRVInstruction *bi, BasicBlock *bb);

  // Translate image gather to LLVM IR
  Value *transSPIRVImageGatherFromInst(SPIRVInstruction *bi, BasicBlock *bb);

  // Translate image fetch/read to LLVM IR
  Value *transSPIRVImageFetchReadFromInst(SPIRVInstruction *bi, BasicBlock *bb);

  // Translate image write to LLVM IR
  Value *transSPIRVImageWriteFromInst(SPIRVInstruction *bi, BasicBlock *bb);

  // Translate OpImageQueryLevels to LLVM IR
  Value *transSPIRVImageQueryLevelsFromInst(SPIRVInstruction *bi, BasicBlock *bb);

  // Translate OpImageQuerySamples to LLVM IR
  Value *transSPIRVImageQuerySamplesFromInst(SPIRVInstruction *bi, BasicBlock *bb);

  // Translate OpImageQuerySize/OpImageQuerySizeLod to LLVM IR
  Value *transSPIRVImageQuerySizeFromInst(SPIRVInstruction *bi, BasicBlock *bb);

  // Translate OpImageQueryLod to LLVM IR
  Value *transSPIRVImageQueryLodFromInst(SPIRVInstruction *bi, BasicBlock *bb);

  // Translate integer dot product to LLVM IR
  Value *transSPIRVIntegerDotProductFromInst(SPIRVInstruction *bi, BasicBlock *bb);

  std::pair<Type *, Value *> createLaunderRowMajorMatrix(Type *const, Value *const);
  Value *addLoadInstRecursively(SPIRVType *const, Value *const, Type *const, bool, bool, bool);
  void addStoreInstRecursively(SPIRVType *const, Value *const, Type *, Value *const, bool, bool, bool);
  Constant *buildConstStoreRecursively(SPIRVType *const, Type *const, Type *const, Constant *const);

  std::pair<Value *, Align> createGepIntoRowMajorMatrix(Type *matrixType, Value *matrixPtr, Align matrixAlign,
                                                        Value *row, Value *col, bool inBounds);

  // Post-process translated LLVM module to undo row major matrices.
  bool postProcessRowMajorMatrix();
  SmallVector<Value *> getTranslatedValues(SPIRVValue *bv, Function *f, BasicBlock *bb);
  Value *getTranslatedValue(SPIRVValue *bv, Function *f, BasicBlock *bb);

  // Create !lgc.xfb.state metadata
  void createXfbMetadata(bool hasXfbOuts);
  bool isRayQueryCommittedIntersection(SPIRVValue *bv);

private:
  class SPIRVTypeContext {
    SPIRVWord m_typeId;
    uint32_t m_matrixStride;
    uint8_t m_predicates;

  public:
    SPIRVTypeContext(SPIRVType *type, uint32_t matrixStride, bool columnMajor, LayoutMode layout)
        : m_typeId(type->getId()), m_matrixStride(matrixStride), m_predicates(0) {
      m_predicates |= uint8_t(columnMajor);
      m_predicates |= uint8_t((uint8_t)layout << 1);
    }

    // Tuple representation to make it easily hashable.
    using TupleType = std::tuple<decltype(m_typeId), decltype(m_matrixStride), decltype(m_predicates)>;
    TupleType asTuple() const { return std::make_tuple(m_typeId, m_matrixStride, m_predicates); }
  };

  typedef DenseMap<SPIRVTypeContext::TupleType, Type *> SPIRVToLLVMFullTypeMap;
  typedef compilerutils::LoweringPointerTupleMap<SPIRVValue *, Value *, true> SPIRVToLLVMValueMap;
  typedef DenseMap<SPIRVValue *, Value *> SPIRVBlockToLLVMStructMap;
  typedef DenseMap<SPIRVFunction *, Function *> SPIRVToLLVMFunctionMap;
  typedef DenseMap<GlobalVariable *, SPIRVBuiltinVariableKind> BuiltinVarMap;
  typedef DenseMap<SPIRVType *, SmallVector<unsigned, 8>> RemappedTypeElementsMap;
  typedef DenseMap<SPIRVValue *, Type *> SPIRVAccessChainValueToLLVMRetTypeMap;
  typedef DenseMap<const SPIRVEntry *, Value *> SPIRVToLLVMEntryMap;
  typedef DenseMap<std::pair<BasicBlock *, BasicBlock *>, unsigned> BlockPredecessorToCountInFunction;

  // A SPIRV value may be translated to a load instruction of a placeholder
  // global variable. This map records load instruction of these placeholders
  // which are supposed to be replaced by the real values later.
  typedef std::map<SPIRVValue *, LoadInst *> SPIRVToLLVMPlaceholderMap;

  Module *m_m;
  BuiltinVarMap m_builtinGvMap;
  LLVMContext *m_context;
  lgc::Builder *m_builder;
  SPIRVModule *m_bm;
  std::string m_globalVarPrefix;
  bool m_enableGatherLodNz;
  ShaderFloatControlFlags m_fpControlFlags;
  ShaderFloatFastMathDefault m_fpFastMathDefault;
  SPIRVFunction *m_entryTarget;
  const SPIRVSpecConstMap &m_specConstMap;
  llvm::ArrayRef<ConvertingSampler> m_convertingSamplers;
  SPIRVToLLVMFullTypeMap m_fullTypeMap;
  SPIRVToLLVMFullTypeMap m_imageTypeMap; // Map to store struct/array with sampler type
  SPIRVToLLVMValueMap m_valueMap;
  SPIRVToLLVMEntryMap m_entryMap;
  SPIRVToLLVMFunctionMap m_funcMap;
  SPIRVBlockToLLVMStructMap m_blockMap;
  SPIRVToLLVMPlaceholderMap m_placeholderMap;
  SPIRVToLLVMDbgTran m_dbgTran;
  DenseMap<std::pair<SPIRVValue *, Function *>, SmallVector<Value *>> m_variableMap;
  DenseMap<SPIRVValue *, Value *> m_variableNonImageMap;
  DenseMap<Type *, Type *> m_multiTypesMap;
  DenseSet<SPIRVType *> m_spvMultiTypesMap;

  // Hash map with correlation between (SPIR-V) OpAccessChain and its returned (dereferenced) type.
  // We have to store base type because opaque-pointers are removing information about dereferenced type.
  SPIRVAccessChainValueToLLVMRetTypeMap m_accessChainRetTypeMap;
  std::map<std::string, unsigned> m_mangleNameToIndex;
  RemappedTypeElementsMap m_remappedTypeElements;
  DenseMap<Type *, bool> m_typesWithPadMap;
  DenseMap<Type *, uint64_t> m_typeToStoreSize;
  DenseMap<std::pair<SPIRVType *, unsigned>, Type *> m_overlappingStructTypeWorkaroundMap;
  DenseMap<Function *, BlockPredecessorToCountInFunction> m_blockPredecessorToCount;
  const Vkgc::ShaderModuleUsage *m_moduleUsage;

  const Vkgc::PipelineShaderOptions *m_shaderOptions;
  bool m_workaroundStorageImageFormats;
  unsigned m_spirvOpMetaKindId;
  unsigned m_execModule;
  bool m_scratchBoundsChecksEnabled;

  bool m_requireFullQuads;

  bool m_maximallyReconverges = false;
  bool m_hasDemoteToHelper = false;
  bool m_hasSamplerInStruct = false;
  enum class LlvmMemOpType : uint8_t { IS_LOAD, IS_STORE };
  struct ScratchBoundsCheckData {
    LlvmMemOpType memOpType;
    SPIRVValue *instructionOrigin;
    SmallVector<llvm::Instruction *, 1> llvmInstructions;
  };

  lgc::xdl::CooperativeMatrixElementType mapToBasicType(SPIRVType *const spvType);
  lgc::xdl::CooperativeMatrixLayout getCooperativeMatrixKHRLayout(CooperativeMatrixUse use,
                                                                  lgc::xdl::CooperativeMatrixElementType elemTy,
                                                                  unsigned rows, unsigned columns);

  Value *transCooperativeMatrixArithInst(SPIRVValue *spvVal, BasicBlock *bb);
  Value *transCooperativeMatrixKHRFromConstruct(SPIRVType *spvCoopMatRowTy, const std::vector<Value *> &constituents);

  // Stores pointers of LLVM Functions to SPIRV memops to the translated LLVM memop(s) in a MapVector to preserve
  // insertion order of the SPIRV memops and to preserve the function origins, as the bounds checks need to be
  // inserted on a per-function level. To handle dependencies between the LLVM IR memops, e. g. using a load result as
  // input for another load, we are using a vector here, which contains a pointer to the instruction along with it being
  // either a load or store.
  MapVector<Function *, MapVector<SPIRVValue *, ScratchBoundsCheckData>> m_spirvMemopToLlvmMemopMapping;
  bool checkVolatile(SPIRVValue *pointer, SPIRVMemoryAccess *access, bool isLoad);
  bool checkCoherent(SPIRVValue *pointer, SPIRVMemoryAccess *access, bool isLoad);

  lgc::Builder *getBuilder() const { return m_builder; }

  // Perform type translation for uncached types. Used in `transType`. Returns the new LLVM type.
  Type *transTypeImpl(SPIRVType *bt, unsigned matrixStride, bool columnMajor, LayoutMode layout);

  Type *getPointeeType(SPIRVValue *v, LayoutMode layout = LayoutMode::Native);

  Type *tryGetAccessChainRetType(SPIRVValue *v) {
    auto loc = m_accessChainRetTypeMap.find(v);
    if (loc != m_accessChainRetTypeMap.end())
      return loc->second;
    return nullptr;
  }

  void tryAddAccessChainRetType(SPIRVValue *v, Type *t) {
    auto loc = m_accessChainRetTypeMap.find(v);
    if (loc == m_accessChainRetTypeMap.end())
      m_accessChainRetTypeMap[v] = t;
  }

  void recordRemappedTypeElements(SPIRVType *bt, unsigned from, unsigned to);

  bool isRemappedTypeElements(SPIRVType *bt) const { return m_remappedTypeElements.count(bt) > 0; }

  unsigned lookupRemappedTypeElements(SPIRVType *bt, unsigned from) {
    assert(m_remappedTypeElements.count(bt) > 0);
    assert(m_remappedTypeElements[bt].size() > from);
    return m_remappedTypeElements[bt][from];
  }

  Type *getPadType(unsigned bytes) { return ArrayType::get(getBuilder()->getInt8Ty(), bytes); }

  void recordTypeWithPad(Type *const t, bool isMatrixRow = false) { m_typesWithPadMap[t] = isMatrixRow; }

  bool isTypeWithPad(Type *const t) const { return m_typesWithPadMap.count(t) > 0; }

  bool isTypeWithPadRowMajorMatrix(Type *const t) const { return m_typesWithPadMap.lookup(t); }

  // Returns a cached type store size. If there is no entry for the given type,
  // its store size is calculated and added to the cache.
  uint64_t getTypeStoreSize(Type *const t);
  // If a value is mapped twice, the existing mapped value is a placeholder,
  // which must be a load instruction of a global variable whose name starts
  // with kPlaceholderPrefix.
  SmallVector<Value *> mapValue(SPIRVValue *bv, ArrayRef<Value *> values);
  SmallVector<Value *> mapValue(SPIRVValue *bv, Value *values);
  // Keep track of the spirvEntry
  Value *mapEntry(const SPIRVEntry *be, Value *v);

  // Used to keep track of the number of incoming edges to a block from each
  // of the predecessor.
  void recordBlockPredecessor(Function *func, BasicBlock *block, BasicBlock *predecessorBlock) {
    assert(func);
    assert(block);
    assert(predecessorBlock);
    m_blockPredecessorToCount[func][{block, predecessorBlock}] += 1;
  }

  unsigned getBlockPredecessorCounts(Function *f, BasicBlock *block, BasicBlock *predecessor);

  bool isSPIRVBuiltinVariable(GlobalVariable *gv, SPIRVBuiltinVariableKind *kind = nullptr);

  // Change this if it is no longer true.
  bool isFuncNoUnwind() const { return true; }

  Value *mapFunction(SPIRVFunction *bf, Function *f) {
    m_funcMap[bf] = f;
    return f;
  }

  SPIRVErrorLog &getErrorLog() { return m_bm->getErrorLog(); }

  void setCallingConv(CallInst *call) {
    Function *f = call->getCalledFunction();
    assert(f);
    call->setCallingConv(f->getCallingConv());
  }

  void setAttrByCalledFunc(CallInst *call);
  Type *transFPType(SPIRVType *t);
  FastMathFlags getFastMathFlags(SPIRVValue *bv);
  void setFastMathFlags(SPIRVValue *bv);
  void setFastMathFlags(Value *val);
  llvm::Value *transBinaryShiftBitwiseInst(SPIRVValue *bv, BasicBlock *bb, Function *f);
  llvm::Value *transCmpInst(SPIRVValue *bv, BasicBlock *bb, Function *f);

  void setName(ArrayRef<Value *> values, SPIRVValue *bv);
  void setLLVMLoopMetadata(SPIRVLoopMerge *lm, BranchInst *bi);
  template <class Source, class Func> bool foreachFuncCtlMask(Source, Func);
  llvm::GlobalValue::LinkageTypes transLinkageType(const SPIRVValue *v);

  void transMemFence(BasicBlock *bb, SPIRVWord memSema, SPIRVWord memScope);
  void truncConstantIndex(std::vector<Value *> &indices, BasicBlock *bb);

  Value *ConvertingSamplerSelectLadderHelper(Value *result, Value *convertingSamplerIdx,
                                             const std::function<Value *(Value *)> &createImageOp);

  bool isStorageClassExplicitlyLaidOut(SPIRVStorageClassKind storageClass);
  bool isStorageClassScalarLayout(SPIRVStorageClassKind storageClass);
  LayoutMode getLayoutModeForStorageClass(SPIRVStorageClassKind storageClass);

  bool hasSpirvType(SPIRVType *spvTy, spv::Op ty);
  Value *createTraceRayDialectOp(SPIRVValue *const spvValue);

  // ========================================================================================================================
  // Wrapper method for easier access to the pipeline context.
  // @returns : Pointer to the pipeline context of the current LLPC context.
  // ========================================================================================================================
  Llpc::PipelineContext *getPipelineContext() const;

  // ========================================================================================================================
  // Wrapper method for easier access to pipeline options.
  // @returns : Pointer to the pipeline options of the current LLPC context.
  // ========================================================================================================================
  const Vkgc::PipelineOptions *getPipelineOptions() const;

  // ========================================================================================================================
  // Wrapper method for easier access to the raytracing context.
  // @returns : Pointer to the pipeline context of the current raytracing context.
  // ========================================================================================================================
  Llpc::RayTracingContext *getRaytracingContext() const;

  // ========================================================================================================================
  // Helper method for checking if the scratch out of bounds check was enabled.
  // @returns : Whether the check is enabled or not.
  // ========================================================================================================================
  bool scratchBoundsChecksEnabled() const;

  // ========================================================================================================================
  // Returns the source access chain (for SPVLoad *) or dest access chain (for SPVStore *) based on a given instruction
  // type.
  // @param [in] memOp: either a SPVLoad or a SPVStore.
  // @param [in] instructionType: an enum value describing the kind of memOp.
  // @returns : The access chain source (for SPVLoad) or access chain dest (for SPVStore).
  // ========================================================================================================================
  SPIRVAccessChainBase *deriveAccessChain(SPIRVValue *memOp, SPIRVToLLVM::LlvmMemOpType instructionType) const;

  // ========================================================================================================================
  // Checks if a given memOp is eligible for inserting the bounds check, e. g. if the access chain is of storage class
  // "private" or "function".
  // @param [in] memOp: either a SPVLoad or a SPVStore.
  // @param [in] instructionType: an enum value describing the kind of memOp.
  // @returns : If the out of bounds check should be inserted for memOp.
  // ========================================================================================================================
  bool shouldInsertScratchBoundsCheck(SPIRVValue *memOp, SPIRVToLLVM::LlvmMemOpType instructionType) const;

  // ========================================================================================================================
  // Helper method that returns the last instruction that was inserted in the current insertion block by the builder.
  // @returns : a pointer to the last LLVM instruction.
  // ========================================================================================================================
  Instruction *getLastInsertedValue();

  // ========================================================================================================================
  // Inserts all instructions that were generated by a call to addLoadInstRecursively / addStoreInstRecursively and
  // inserts them in the m_spirvMemopToLlvmMemopMapping map with some additional data.
  // @param [in] spirvMemOp : the key part for the map.
  // @param [in] baseNode : the last instruction inserted before an add*InstRecursively call.
  // @param [in] baseNodeParent: baseNode's basic block.
  // @param [in] instructionType: describes if spirvMemOp is either a SPVLoad or a SPVStore.
  // @param [in] parent: baseNodeParent's function. Also used as key for the m_spirvMemopToLlvmMemopMapping.
  // ========================================================================================================================
  void gatherScratchBoundsChecksMemOps(SPIRVValue *spirvMemOp, Instruction *baseNode, BasicBlock *baseNodeParent,
                                       SPIRVToLLVM::LlvmMemOpType instructionType, Function *parent);

  // ========================================================================================================================
  // Goes through all recorded SPIRV memory instructions and wraps the corresponding LLVM instructions with out of
  // bounds guards against the SPIRV access chain indices. This moves some LLVM instructions in new basic blocks. This
  // also manipulates m_valueMap entries in case memOp is a SPVLoad.
  // @param [in] memOp : the SPIRV memory instruction whose access chain indices should be checked for out of bounds
  // accesses.
  // @param [in] scratchBoundsCheckData : information about the type of SPIRV memory instruction, all LLVM instructions
  // that were created during translation of memOp and additional information.
  // @param [in] parent : the LLVM function that is currently translated.
  // ========================================================================================================================
  void insertScratchBoundsChecks(SPIRVValue *const memOp, const ScratchBoundsCheckData &scratchBoundsCheckData,
                                 Function *parent);

}; // class SPIRVToLLVM

} // namespace SPIRV

#endif
